"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSParser = exports.TSCase = void 0;
const types_1 = require("./types");
const commons_1 = require("./commons");
const data_1 = require("./data");
const strings_1 = require("./strings");
const tscolor_1 = require("./tscolor");
const tsdata_1 = require("./tsdata");
const tsdate_1 = require("./tsdate");
const tsphonenumber_1 = require("./tsphonenumber");
const utils_1 = require("./utils");
const tscountry_1 = require("./tscountry");
const tscharset_1 = require("./tscharset");
var TSCase;
(function (TSCase) {
    TSCase["standard"] = "standard";
    TSCase["lowercase"] = "lowercase";
    TSCase["uppercase"] = "uppercase";
})(TSCase = exports.TSCase || (exports.TSCase = {}));
class TSParser {
    static define(rootNode, errors) {
        return (0, commons_1.$valueornull)(_structureConstruction(rootNode, errors));
    }
    get isValid() { return this.errors.length === 0; }
    validate(value, errors) {
        return this._validate(value, '', errors);
    }
    interpret(value, errors) {
        return this._validate(value, '', errors) ? this.rawInterpret(value) : null;
    }
    encode(value, errors) {
        return this._validate(value, '', errors) ? this.rawEncode(value) : null;
    }
    parse(source, errors) {
        let value = undefined;
        const s = (0, strings_1.$trim)(source);
        if (s.length) {
            try {
                value = JSON.parse(s);
            }
            catch (_a) {
                value = undefined;
                _serror(errors, 'JSON', '.parse(stringSource) did fail');
            }
        }
        else {
            _serror(errors, 'stringSource', ' is empty');
        }
        return this.interpret(value, errors);
    }
    stringify(value, spaces = undefined, errors) {
        return this._validate(value, '', errors) ? this.rawStringify(value, spaces) : null;
    }
    rawStringify(value, spaces = undefined) {
        try {
            return JSON.stringify(this.rawEncode(value), undefined, spaces);
        }
        catch (_a) {
            return null;
        }
    }
    toString() { return (0, utils_1.$inspect)(this.toJSON(), 15); }
    _validate(value, path, errors) {
        if (this.mandatory && !(0, commons_1.$ok)(value)) {
            return _serror(errors, path, 'is mandatory');
        }
        return true;
    }
    constructor(mandatory, checker, transformer, natifier) {
        this.errors = [];
        this._check = (_) => true;
        this._transform = (v) => v;
        this._natify = undefined;
        this.mandatory = mandatory;
        if ((0, commons_1.$isfunction)(checker)) {
            this._check = checker;
        }
        if ((0, commons_1.$isfunction)(transformer)) {
            this._transform = transformer;
        }
        if ((0, commons_1.$isfunction)(natifier)) {
            this._natify = natifier;
        }
    }
    ;
}
exports.TSParser = TSParser;
class TSDictionaryParser extends TSParser {
    constructor(node, keysParser, valuesParser) {
        const m = node._mandatory;
        super(!!m, node._checker, node._transformer, node._natifier);
        if (!(keysParser instanceof TSLeafParser) || !keysParser.isKey) {
            this.errors.push('_keysType field is not a valid node key definition');
        }
        if ((0, commons_1.$defined)(m) && !(0, commons_1.$isbool)(m)) {
            this.errors.push(`_mandatory field should be a boolean`);
        }
        if ((0, commons_1.$ok)(node._keysCase)) {
            this.errors.push(`_keysCase field should not be set for a dictionary`);
        }
        this._keysParser = keysParser;
        this._valuesParser = valuesParser;
    }
    toJSON() {
        return {
            _type: 'dictionary',
            _mandatory: this.mandatory,
            _check: this._check,
            _transform: this._transform,
            _natify: this._natify,
            _keysStruct: this._keysParser.toJSON(),
            _valuesStruct: this._valuesParser.toJSON()
        };
    }
    _validate(value, path, errors) {
        if (!super._validate(value, path, errors)) {
            return false;
        }
        if ((0, commons_1.$ok)(value) && ((0, commons_1.$isarray)(value) || !(0, commons_1.$isobject)(value))) {
            return _serror(errors, path, 'is not a dictionary');
        }
        const entries = Object.entries(value);
        let founds = 0;
        const p = path.length ? path : 'dict';
        for (let [k, v] of entries) {
            if (!this._keysParser._validate(k, `${p}.${k}[key]`, errors))
                return false;
            if (!this._valuesParser._validate(v, `${p}.${k}`, errors))
                return false;
            founds++;
        }
        if (founds === 0 && this.mandatory) {
            return _serror(errors, path, ` is empty`);
        }
        return this._check(value) ? true : _serror(errors, path, 'specific check did fail');
    }
    rawInterpret(value) {
        const ret = {};
        const entries = Object.entries(value);
        for (let [k, v] of entries) {
            const ki = this._keysParser.rawEncode(k);
            const vi = this._valuesParser.rawInterpret(v);
            if ((0, commons_1.$ok)(ki) && (0, commons_1.$ok)(vi)) {
                ret[ki] = vi;
            }
        }
        return this._transform(ret);
    }
    rawEncode(value) {
        const ret = {};
        const entries = Object.entries(value);
        for (let [k, v] of entries) {
            const ki = this._keysParser.rawEncode(k);
            const vi = this._valuesParser.rawEncode(v);
            if ((0, commons_1.$ok)(ki) && (0, commons_1.$ok)(vi)) {
                ret[ki] = vi;
            }
        }
        return this._natify ? this._natify(ret) : ret;
    }
}
class TSObjectParser extends TSParser {
    constructor(node, itemsParsers) {
        const m = node._mandatory;
        super(!!m, node._checker, node._transformer, node._natifier);
        this._keyTransform = (s) => s;
        this._keysCase = TSCase.standard;
        if ((0, commons_1.$defined)(m) && !(0, commons_1.$isbool)(m)) {
            this.errors.push(`_mandatory field should be a boolean`);
        }
        const l = (0, commons_1.$value)(node._keysCase, TSCase.standard);
        const kt = InternalCasingMap[l];
        if (!(0, commons_1.$defined)(kt)) {
            this.errors.push(`_keysCase field should one of ${Array.from((0, commons_1.$keys)(InternalCasingMap))}`);
        }
        else {
            this._keyTransform = kt;
            this._keysCase = l;
        }
        this._itemsParsers = itemsParsers;
        this._count = (0, commons_1.$objectcount)(itemsParsers);
    }
    toJSON() {
        const ret = {
            _type: 'object',
            _mandatory: this.mandatory,
            _check: this._check,
            _transform: this._transform,
            _natify: this._natify
        };
        if (this._keysCase !== 'standard') {
            ret._keysCase = this._keysCase;
        }
        const structEntries = Object.entries(this._itemsParsers);
        for (let [k, struct] of structEntries) {
            ret[k] = struct.toJSON();
        }
        return ret;
    }
    _validate(value, path, errors) {
        if (!super._validate(value, path, errors)) {
            return false;
        }
        if ((0, commons_1.$ok)(value) && ((0, commons_1.$isarray)(value) || !(0, commons_1.$isobject)(value))) {
            return _serror(errors, path, 'is not an object');
        }
        const p = path.length ? path : 'object';
        const entries = Object.entries(value);
        const founds = new Set();
        for (let [key, v] of entries) {
            const k = this._keyTransform(key);
            if (founds.has(k)) {
                return _serror(errors, path, `.${k} is present several times`);
            }
            const struct = this._itemsParsers[k];
            if (!(0, commons_1.$ok)(struct)) {
                return _serror(errors, path, `.${k} is unknown`);
            }
            if (!struct._validate(v, `${p}.${k}`, errors))
                return false;
            founds.add(k);
        }
        if (founds.size < this._count) {
            const structEntries = Object.entries(this._itemsParsers);
            for (let [k, struct] of structEntries) {
                if (!founds.has(k) && struct.mandatory) {
                    return _serror(errors, path, `.${k} is mandatory`);
                }
            }
        }
        return this._check(value) ? true : _serror(errors, path, 'specific check did fail');
    }
    rawInterpret(value) {
        const ret = {};
        const entries = Object.entries(value);
        for (let [key, v] of entries) {
            const k = this._keyTransform(key);
            const struct = this._itemsParsers[k];
            const res = struct.rawInterpret(v);
            if ((0, commons_1.$ok)(res)) {
                ret[k] = res;
            }
        }
        return this._transform(ret);
    }
    rawEncode(value) {
        const ret = {};
        const entries = Object.entries(value);
        for (let [key, v] of entries) {
            const k = this._keyTransform(key);
            const struct = this._itemsParsers[k];
            const res = struct.rawEncode(v);
            if ((0, commons_1.$ok)(res)) {
                ret[k] = res;
            }
        }
        return this._natify ? this._natify(ret) : ret;
    }
}
class TSArrayParser extends TSParser {
    constructor(node, itemParser) {
        const m = node._mandatory;
        let min = 0, max = types_1.UINT_MAX;
        const errors = [];
        if ((0, commons_1.$ok)(node._min)) {
            if (!(0, commons_1.$isunsigned)(node._min)) {
                errors.push(`wrong minimal size definition for array node`);
            }
            else {
                min = (0, commons_1.$unsigned)(node._min);
            }
        }
        if ((0, commons_1.$ok)(node._max)) {
            if (!(0, commons_1.$isunsigned)(node._max)) {
                errors.push(`wrong maximal size definition for array node`);
            }
            else {
                max = (0, commons_1.$unsigned)(node._max);
            }
        }
        if (min > max) {
            errors.push(`maximal size < minimal size for array node`);
        }
        super((0, commons_1.$ok)(m) ? !!m : min > 0, node._checker, node._transformer, node._natifier);
        if ((0, commons_1.$defined)(m) && !(0, commons_1.$isbool)(m)) {
            this.errors.push(`_mandatory field should be a boolean`);
        }
        this._min = min;
        this._max = max;
        this._itemsParser = itemParser;
        for (let e of errors) {
            this.errors.push(e);
        }
    }
    toJSON() {
        return {
            _type: 'array',
            _mandatory: this.mandatory,
            _min: this._min > 0 ? this._min : 'none',
            _max: this._max < types_1.UINT_MAX ? this._max : 'none',
            _check: this._check,
            _transform: this._transform,
            _natify: this._natify,
            _itemsStruct: this._itemsParser.toJSON()
        };
    }
    _validate(value, path, errors) {
        if (!super._validate(value, path, errors)) {
            return false;
        }
        if ((0, commons_1.$ok)(value) && !(0, commons_1.$isarray)(value)) {
            return _serror(errors, path, 'is not an array');
        }
        const n = (0, commons_1.$count)(value);
        if (n < this._min) {
            return _serror(errors, path, `.size is too small (${n} < ${this._min})`);
        }
        if (n > this._max) {
            return _serror(errors, path, `.size is too large (${n} > ${this._max})`);
        }
        const p = path.length ? path : 'array';
        for (let i = 0; i < n; i++) {
            if (!this._itemsParser._validate(value[i], `${p}[${i}]`, errors))
                return false;
        }
        return this._check(value) ? true : _serror(errors, path, 'specific check did fail');
    }
    rawInterpret(value) {
        const ret = [];
        for (let v of value) {
            ret.push(this._itemsParser.rawInterpret(v));
        }
        return this._transform(ret);
    }
    rawEncode(value) {
        const ret = [];
        for (let v of value) {
            ret.push(this._itemsParser.rawEncode(v));
        }
        return this._natify ? this._natify(ret) : ret;
    }
}
class TSLeafParser extends TSParser {
    constructor(node) {
        const manager = (0, commons_1.$isstring)(node._type) ? TSLeafParser.__managers[node._type] : undefined;
        const transformer = (0, commons_1.$ok)(node._transformer) ? node._transformer : ((0, commons_1.$ok)(manager === null || manager === void 0 ? void 0 : manager.trans) ? manager.trans : undefined);
        const natifier = (0, commons_1.$ok)(node._natifier) ? node._natifier : ((0, commons_1.$ok)(manager === null || manager === void 0 ? void 0 : manager.v2nat) ? manager.v2nat : undefined);
        super(!!node._mandatory, node._checker, transformer, natifier);
        if (!(0, commons_1.$ok)(manager)) {
            this.errors.push(`Invalid parser type '${node._type}'`);
            this._manager = TSLeafParser.__dummyManager;
        }
        else {
            this._manager = manager;
        }
        const enumeration = node._enum;
        if ((0, commons_1.$isarray)(enumeration)) {
            if (!(0, commons_1.$ok)(manager.enum)) {
                this.errors.push(`Parser type '${node._type}' does not support enumeration`);
            }
            if (!enumeration.length) {
                this.errors.push(`Empty enumeration definition array for type '${node._type}'`);
            }
            for (let e of enumeration) {
                if (!manager.enum(e)) {
                    this.errors.push(`Enumeration value '${e}' is invalid for type '${node._type}'`);
                }
            }
            this._enumeration = new Set(enumeration);
        }
        else if ((0, commons_1.$isobject)(enumeration)) {
            if (!(0, commons_1.$ok)(manager.enum)) {
                this.errors.push(`Parser type '${node._type}' does not support enumeration`);
            }
            const entries = Object.entries(enumeration);
            this._enumeration = new Set();
            for (let [key, item] of entries) {
                if (!key.length || !key.match(IsFielOrEnumdRegex)) {
                    this.errors.push(`Wrong enumeration key '${key}' for type '${node._type}'`);
                }
                if (manager.enum(item)) {
                    this._enumeration.add(item);
                }
                else {
                    this.errors.push(`Enumeration value '${item}' for key '${key}' is invalid for type '${node._type}'`);
                }
            }
            if (!this._enumeration.size) {
                this.errors.push(`Empty enumeration definition dictionary for type '${node._type}'`);
            }
            this._conversion = enumeration;
        }
        else if ((0, commons_1.$ok)(enumeration)) {
            this.errors.push(`Bad enumeration definition for type '${node._type}'`);
        }
        this._type = node._type;
    }
    get isKey() { return !!this._manager.iskey; }
    toJSON() {
        return {
            _type: this._type,
            _mandatory: this.mandatory,
            _enumeration: (0, commons_1.$ok)(this._enumeration) ? Array.from(this._enumeration) : 'none',
            _conversion: (0, commons_1.$value)(this._conversion, 'none'),
            _check: this._check,
            _transform: this._transform,
            _natify: this._natify
        };
    }
    _validate(value, path, errors) {
        if (!super._validate(value, path, errors)) {
            return false;
        }
        if (this._conversion && (0, commons_1.$isstring)(value)) {
            const tag = value;
            value = this._conversion[tag];
            if (!(0, commons_1.$ok)(value)) {
                return _serror(errors, path, `has not a valid enum tag ${tag}`);
            }
        }
        if ((0, commons_1.$ok)(value) && !this._manager.valid(value)) {
            return _serror(errors, path, `is not a valid ${this._type}`);
        }
        if (this._enumeration) {
            if (!this._enumeration.has(this._convertIfString(value))) {
                const s = (0, commons_1.$isstring)(value) ? "'" + value + "'" : value;
                return _serror(errors, path, `has not a valid enum value ${s}`);
            }
        }
        return this._check(value) ? true : _serror(errors, path, 'specific check did fail');
    }
    _convertIfString(value) { return this._transform((0, commons_1.$isstring)(value) ? this._manager.str2v(value) : value); }
    rawInterpret(value) {
        if (this._conversion && (0, commons_1.$isstring)(value)) {
            value = this._conversion[value];
        }
        return this._convertIfString(value);
    }
    rawEncode(value) {
        if (this._conversion && (0, commons_1.$isstring)(value)) {
            value = this._conversion[value];
        }
        if (!(0, commons_1.$ok)(value)) {
            return value;
        }
        switch (typeof value) {
            case 'bigint': return Number(value);
            case 'boolean': return value;
            case 'number': return value;
            case 'object': return this._natify ? this._natify(value) : (0, commons_1.$string)(value);
            case 'string': return value;
            case 'symbol': return (0, commons_1.$string)(value);
            default: return null;
        }
    }
}
TSLeafParser.__managers = {
    'boolean': { valid: _isBoolean, str2v: (s) => !!_booleanConvertor(s) },
    'charset': { valid: _isCharset, str2v: (s) => tscharset_1.TSCharset.charset(s), v2nat: (v) => v.name },
    'color': { valid: _iscolor, str2v: (s) => tscolor_1.TSColor.fromString(s), v2nat: _color2str },
    'continent': { valid: _isContinent, str2v: (s) => s, enum: _isContinent, iskey: true },
    'country': { valid: _isCountry, trans: _countryTrans, str2v: (s) => s, v2nat: (v) => v.alpha2Code, enum: (v) => types_1.TSCountrySet.has(v), iskey: true },
    'currency': { valid: _isCurrency, str2v: (s) => s, enum: _isCurrency, iskey: true },
    'data': { valid: _isData, str2v: data_1.$decodeBase64, v2nat: data_1.$encodeBase64 },
    'date': { valid: _isTsDate, str2v: (s) => tsdate_1.TSDate.fromIsoString(s), v2nat: (v) => v.toIsoString() },
    'email': { valid: commons_1.$isemail, str2v: commons_1.$email, iskey: true },
    'int': { valid: _isInt, str2v: commons_1.$int, enum: commons_1.$isint, iskey: true },
    'int8': { valid: (v) => _isInt(v, types_1.INT8_MIN, types_1.INT8_MAX), str2v: commons_1.$int, enum: (v) => (0, commons_1.$isint)(v, types_1.INT8_MIN, types_1.INT8_MAX), iskey: true },
    'int16': { valid: (v) => _isInt(v, types_1.INT16_MIN, types_1.INT16_MAX), str2v: commons_1.$int, enum: (v) => (0, commons_1.$isint)(v, types_1.INT16_MIN, types_1.INT16_MAX), iskey: true },
    'int32': { valid: (v) => _isInt(v, types_1.INT32_MIN, types_1.INT32_MAX), str2v: commons_1.$int, enum: (v) => (0, commons_1.$isint)(v, types_1.INT32_MIN, types_1.INT32_MAX), iskey: true },
    'jsdate': { valid: _isJsDate, str2v: (s) => new Date(s), v2nat: (v) => v.toISOString },
    'language': { valid: _isLanguage, str2v: (s) => s, enum: _isLanguage, iskey: true },
    'phone': { valid: commons_1.$isphonenumber, str2v: (s) => tsphonenumber_1.TSPhoneNumber.fromString(s), v2nat: (v) => v.standardNumber, iskey: true },
    'number': { valid: _isNumber, str2v: (s) => Number(s), enum: commons_1.$isnumber },
    'string': { valid: commons_1.$isstring, str2v: (s) => s, enum: (v) => (0, commons_1.$isstring)(v) && v.length > 0 },
    'uint8': { valid: (v) => _isInt(v, 0, types_1.UINT8_MAX), str2v: commons_1.$unsigned, enum: (v) => (0, commons_1.$isunsigned)(v, types_1.UINT8_MAX), iskey: true },
    'uint16': { valid: (v) => _isInt(v, 0, types_1.UINT16_MAX), str2v: commons_1.$unsigned, enum: (v) => (0, commons_1.$isunsigned)(v, types_1.UINT16_MAX), iskey: true },
    'uint32': { valid: (v) => _isInt(v, 0, types_1.UINT32_MAX), str2v: commons_1.$unsigned, enum: (v) => (0, commons_1.$isunsigned)(v, types_1.UINT32_MAX), iskey: true },
    'unsigned': { valid: (v) => _isInt(v, 0), str2v: commons_1.$unsigned, enum: commons_1.$isunsigned, iskey: true },
    'url': { valid: commons_1.$isurl, str2v: commons_1.$url, iskey: true },
    'uuid': { valid: commons_1.$isuuid, str2v: commons_1.$UUID, iskey: true },
};
TSLeafParser.__dummyManager = { valid: (_) => false, str2v: (_) => undefined };
function _isCharset(v) { return v instanceof tscharset_1.TSCharset || ((0, commons_1.$isstring)(v) && (0, commons_1.$ok)(tscharset_1.TSCharset.charset(v))); }
function _isNumber(v) { return (0, commons_1.$isnumber)(v) || ((0, commons_1.$isstring)(v) && (0, commons_1.$isnumber)(Number(v))); }
function _isData(v) { return v instanceof tsdata_1.TSData || v instanceof Uint8Array || v instanceof ArrayBuffer || (0, commons_1.$isstring)(v); }
function _isBoolean(v) { return (0, commons_1.$isbool)(v) || ((0, commons_1.$isstring)(v) && (0, commons_1.$ok)(_booleanConvertor(v))); }
function _isContinent(v) { return types_1.TSContinentSet.has(v); }
function _isCurrency(v) { return types_1.TSCurrencySet.has(v); }
function _isLanguage(v) { return types_1.TSLanguageSet.has(v); }
function _isCountry(v) { return v instanceof tscountry_1.TSCountry || types_1.TSCountrySet.has(v); }
function _countryTrans(v) { return v instanceof tscountry_1.TSCountry ? v.alpha2Code : v; }
function _isInt(v, min = types_1.INT_MIN, max = types_1.INT_MAX) { return (0, commons_1.$isint)(v, min, max) || ((0, commons_1.$isstring)(v) && (0, commons_1.$isint)(Number(v), min, max)); }
function _isJsDate(v) { return v instanceof Date || ((0, commons_1.$isstring)(v) && v !== '0' && (0, commons_1.$isnumber)(Date.parse(v))); }
function _isTsDate(v) { return v instanceof tsdate_1.TSDate || ((0, commons_1.$isstring)(v) && (0, commons_1.$ok)((0, commons_1.$isodate)(v))); }
function _iscolor(o) { return o instanceof tscolor_1.TSColor || ((0, commons_1.$isstring)(o)) && (0, commons_1.$ok)(tscolor_1.TSColor.fromString(o)); }
function _booleanConvertor(s) {
    s = (0, strings_1.$ascii)((0, strings_1.$ftrim)(s)).toLowerCase();
    return s === 'true' ? true : (s === 'false' ? false : null);
}
function _color2str(v) { return v.toString({ colorSpace: tscolor_1.TSColorSpace.RGB, removeAlpha: false, rgbaCSSLike: true, shortestCSS: false, uppercase: false }); }
function _serror(errors, path, error) {
    errors === null || errors === void 0 ? void 0 : errors.push(`${path.length > 0 ? path : 'value'}${error.startsWith('.') ? '' : ' '}${error}${error.endsWith('') ? '' : '.'}`);
    return false;
}
const IsFielOrEnumdRegex = /^[a-zA-Z][a-zA-Z0-9_-]*$/g;
const InternalFieldsSet = new Set(['_mandatory', '_keysCase', '_checker', '_transformer', '_natifier']);
const InternalCasingMap = {
    'standard': (s) => s,
    'lowercase': (s) => s.toLowerCase(),
    'uppercase': (s) => s.toUpperCase()
};
function _structureConstruction(node, errors) {
    if ((0, commons_1.$isstring)(node)) {
        let mandatory = node.endsWith('!');
        if (mandatory) {
            node = node.left(node.length - 1);
        }
        return new TSLeafParser({ _type: node, _mandatory: mandatory });
    }
    else if ((0, commons_1.$isarray)(node)) {
        const a = node;
        if (a.length === 0 || a.length > 3) {
            errors === null || errors === void 0 ? void 0 : errors.push('wrong empty array node definition');
            return null;
        }
        let min = undefined, max = undefined;
        if (a.length > 1 && (0, commons_1.$ok)(a[1])) {
            if (!(0, commons_1.$isunsigned)(a[1])) {
                errors === null || errors === void 0 ? void 0 : errors.push(`wrong minimal size definition for array node`);
            }
            else {
                min = (0, commons_1.$unsigned)(a[1]);
            }
        }
        if (a.length > 2 && (0, commons_1.$ok)(a[2])) {
            if (!(0, commons_1.$isunsigned)(a[2])) {
                errors === null || errors === void 0 ? void 0 : errors.push(`wrong maximal size definition for array node`);
            }
            else {
                max = (0, commons_1.$unsigned)(a[2]);
            }
        }
        const itemsType = a[0];
        const itemsParser = _structureConstruction(itemsType, errors);
        if (!(itemsParser === null || itemsParser === void 0 ? void 0 : itemsParser.isValid)) {
            if ((0, commons_1.$count)(itemsParser === null || itemsParser === void 0 ? void 0 : itemsParser.errors)) {
                errors === null || errors === void 0 ? void 0 : errors.push(...itemsParser.errors);
            }
            return null;
        }
        return new TSArrayParser({ _itemsType: itemsType, _min: min, _max: max }, itemsParser);
    }
    else if ((0, commons_1.$isobject)(node)) {
        if ((0, commons_1.$ok)(node._type)) {
            return new TSLeafParser(node);
        }
        else if ((0, commons_1.$ok)(node._keysType) || (0, commons_1.$ok)(node._valueItemsType)) {
            if (!(0, commons_1.$ok)(node._keysType)) {
                errors === null || errors === void 0 ? void 0 : errors.push('_keysType field is not set for dictionary node definition');
            }
            const keysParser = _structureConstruction(node._keysType, errors);
            if (!(keysParser === null || keysParser === void 0 ? void 0 : keysParser.isValid)) {
                if ((0, commons_1.$count)(keysParser === null || keysParser === void 0 ? void 0 : keysParser.errors)) {
                    errors === null || errors === void 0 ? void 0 : errors.push(...keysParser.errors);
                }
                return null;
            }
            if (!(0, commons_1.$ok)(node._valueItemsType)) {
                errors === null || errors === void 0 ? void 0 : errors.push('_valueItemsType field is not set for a dictionary node definition');
            }
            const itemsParser = _structureConstruction(node._valueItemsType, errors);
            if (!(itemsParser === null || itemsParser === void 0 ? void 0 : itemsParser.isValid)) {
                if ((0, commons_1.$count)(itemsParser === null || itemsParser === void 0 ? void 0 : itemsParser.errors)) {
                    errors === null || errors === void 0 ? void 0 : errors.push(...itemsParser.errors);
                }
                return null;
            }
            return new TSDictionaryParser(node, keysParser, itemsParser);
        }
        else if ((0, commons_1.$ok)(node._itemsType)) {
            const itemsParser = _structureConstruction(node._itemsType, errors);
            if (!(itemsParser === null || itemsParser === void 0 ? void 0 : itemsParser.isValid)) {
                if ((0, commons_1.$count)(itemsParser === null || itemsParser === void 0 ? void 0 : itemsParser.errors)) {
                    errors === null || errors === void 0 ? void 0 : errors.push(...itemsParser.errors);
                }
                return null;
            }
            return new TSArrayParser(node, itemsParser);
        }
        else {
            const itemsParsers = {};
            const entries = Object.entries(node);
            const keyscase = (0, commons_1.$value)(InternalCasingMap[(0, commons_1.$value)(node._keysCase, TSCase.standard)], (s) => s);
            let found = 0;
            for (let [key, subnode] of entries) {
                if (!InternalFieldsSet.has(key)) {
                    if (!key.match(IsFielOrEnumdRegex)) {
                        errors === null || errors === void 0 ? void 0 : errors.push(`wrong key '${key}' for object node definition`);
                    }
                    else {
                        const itemParser = _structureConstruction(subnode, errors);
                        if (!(itemParser === null || itemParser === void 0 ? void 0 : itemParser.isValid)) {
                            if ((0, commons_1.$count)(itemParser === null || itemParser === void 0 ? void 0 : itemParser.errors)) {
                                errors === null || errors === void 0 ? void 0 : errors.push(...itemParser.errors);
                            }
                            return null;
                        }
                        itemsParsers[keyscase(key)] = itemParser;
                        found++;
                    }
                }
            }
            if (!found) {
                errors === null || errors === void 0 ? void 0 : errors.push('empty object node definition');
                return null;
            }
            return new TSObjectParser(node, itemsParsers);
        }
    }
    return null;
}
//# sourceMappingURL=tsparser.js.map